/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.cassandra.gms;

import java.io.*;

import com.google.common.annotations.VisibleForTesting;

import org.apache.cassandra.db.TypeSizes;
import org.apache.cassandra.io.IVersionedSerializer;
import org.apache.cassandra.io.util.DataInputPlus;
import org.apache.cassandra.io.util.DataOutputPlus;

/**
 * HeartBeat State associated with any given endpoint.
 */

public class HeartBeatState
{
    public static final int EMPTY_VERSION = -1;

    public static final IVersionedSerializer<HeartBeatState> serializer = new HeartBeatStateSerializer();

    private volatile int generation;
    private volatile int version;

    HeartBeatState(int gen)
    {
        this(gen, 0);
    }

    public HeartBeatState(HeartBeatState other)
    {
        generation = other.generation;
        version = other.version;
    }

    public HeartBeatState(int gen, int ver)
    {
        generation = gen;
        version = ver;
    }

    public static HeartBeatState empty()
    {
        return new HeartBeatState(0, EMPTY_VERSION);
    }

    public boolean isEmpty()
    {
        // Instance I1 will update this value for I1's state, no other instance should.
        // It is also known that negative version isn't allowed, so can leverage this to
        // know that the state was not generated by I1 but rather than instance (normally
        // happens on startup, the peers are added to gossip with the empty state).
        return version == EMPTY_VERSION;
    }

    public int getGeneration()
    {
        return generation;
    }

    void updateHeartBeat()
    {
        version = VersionGenerator.getNextVersion();
    }

    public int getHeartBeatVersion()
    {
        return version;
    }

    void forceNewerGenerationUnsafe()
    {
        generation += 1;
    }

    @VisibleForTesting
    public void forceHighestPossibleVersionUnsafe()
    {
        version = Integer.MAX_VALUE;
    }

    public String toString()
    {
        return String.format("HeartBeat: generation = %d, version = %d", generation, version);
    }
}

class HeartBeatStateSerializer implements IVersionedSerializer<HeartBeatState>
{
    public void serialize(HeartBeatState hbState, DataOutputPlus out, int version) throws IOException
    {
        out.writeInt(hbState.getGeneration());
        out.writeInt(hbState.getHeartBeatVersion());
    }

    public HeartBeatState deserialize(DataInputPlus in, int version) throws IOException
    {
        return new HeartBeatState(in.readInt(), in.readInt());
    }

    public long serializedSize(HeartBeatState state, int version)
    {
        return TypeSizes.sizeof(state.getGeneration()) + TypeSizes.sizeof(state.getHeartBeatVersion());
    }
}
